/***************************************************************************
                          planet.cpp  -  description
                             -------------------
    begin                : Wed Apr 3 2002
    copyright            : (C) 2002 by Michael Andreen
    email                : whale@linux.nu
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "planet.h"

#include <iostream>
using namespace std;

Planet::Planet()
{
	m_sRace = "Planet";
	m_iStays = -1;
}

Planet::~Planet(){
}

//////////////////////////////////////////////////////////////////////////
//
unsigned Planet::planetScore(int tick) const
{
	int ticks = m_Score.size();
	if (ticks == 0)
		return 0;

	--ticks;

	if (ticks < tick)
		return m_Score[ticks];

	return m_Score[tick];
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::setPlanetScore(unsigned number, int tick)
{
	int earlier = 0;
	int ticks = m_Score.size();

	if (ticks != 0)
		earlier = m_Score[ticks - 1];

	for (int i = ticks; i <= tick; ++i)
	{
		m_Score.push_back(earlier);
	}
	m_Score[tick] = number;
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::addPlanetScore(unsigned number, int tick)
{
	int earlier = 0;
	int ticks = m_Score.size();

	if (ticks != 0)
		earlier = m_Score[ticks - 1];

	for (int i = ticks; i <= tick; ++i)
	{
		m_Score.push_back(earlier);
	}
	m_Score[tick] += number;
}
//////////////////////////////////////////////////////////////////////////
//
int Planet::roids(std::string type, int tick) const
{
	// const... I would like [] as for const types: int ticks = m_Roids[type].size();
	
	vector<int>const* roids = 0;
	for (RoidList::const_iterator i = m_Roids.begin(); i != m_Roids.end(); ++i)
	{
		if (i->first == type)
		{
			roids = &i->second;
			break;
		}
	}
	if (roids == 0)
		return 0;

	int ticks = roids->size();

	if( ticks == 0)
		return 0;
	--ticks;
	if (ticks < tick)
		return roids->at(ticks);
	return roids->at(tick);
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::setRoids(std::string type, int number, int tick)
{
	int ticks = m_Roids[type].size();
	int roids = 0;
	if (m_Roids[type].size() > 0)
		roids = m_Roids[type][m_Roids[type].size() - 1];

	for (int i = ticks; i <= tick; ++i )
		m_Roids[type].push_back(roids);
	m_Roids[type][tick] = number;
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::takeRoids(std::string type, int number, int tick)
{
	int ticks = m_Roids[type].size();
	int roids = 0;
	if (m_Roids[type].size() > 0)
		roids = m_Roids[type][m_Roids[type].size() - 1];

	for (int i = ticks; i <= tick; ++i )
		m_Roids[type].push_back(roids);
	m_Roids[type][tick] -= number;

	if (type != "uninit")
		addPlanetScore(-1000*number, tick);
}



//////////////////////////////////////////////////////////////////////////
//
void Planet::runBattle(std::vector<Fleet*> friendly, std::vector<Fleet*> hostile)
{
	if (hostile.size() == 0)
			return;

	int skipped = 0;

	for (vector<Fleet*>::iterator i = friendly.begin(); i != friendly.end(); ++i)
		(*i)->resetTicks();

	for (vector<Fleet*>::iterator i = hostile.begin(); i != hostile.end(); ++i)
		(*i)->resetTicks();

	m_Report.clear();

	for(int tick = 1; skipped < 20; ++tick)
	{
		//See who's in the battle at the current tick
		vector<Fleet*> friends = calculateSide(friendly, tick);
		vector<Fleet*> hostiles = calculateSide(hostile, tick);

		// No idea to calculate anything if noone is there.. ;)
		if (hostiles.size() == 0)
		{
			skipped++;
			continue;
		}
		else
			skipped = 0;

		Planet allFriends;
		allFriends.addToThis(friends, tick - 1);

		/*for (RoidList::iterator i = m_Roids.begin(); i != m_Roids.end(); ++i)
			allFriends.setRoids(i->first, roids(i->first, tick));

		allFriends.setPlanetScore(m_iScore);*/
		
		Fleet allHostiles;
		allHostiles.addToThis(hostiles, tick - 1);

		map<string, map<string, int> > stealfriendly;
		map<string, map<string, int> > stealhostile;

		//Reset roids
		for (RoidList::iterator roid = m_Roids.begin(); roid != m_Roids.end(); ++roid)
		{
			setRoids(roid->first, roids(roid->first, tick-1), tick);
		}

		setPlanetScore(planetScore(tick - 1), tick);
		
		
		calcOneTick(&allFriends, &allHostiles, stealfriendly, stealhostile, tick );

		//allFriends.printFleet();

		allFriends.distributeLossesGains(friends, tick);
		allFriends.distributeStolenShips(stealfriendly, friends, tick);

		allHostiles.distributeLossesGains(hostiles, tick);
		allHostiles.distributeCappedRoids(hostiles, tick);
		allHostiles.distributeStolenShips(stealhostile, hostiles, tick);

		calculateScoreLoss(friends, tick);

	/*	for (RoidList::iterator i = m_Roids.begin(); i != m_Roids.end(); ++i)
			setRoids(i->first, roids(i->first, 1), tick);*/
	}

	for (vector<Fleet*>::iterator i = friendly.begin(); i != friendly.end(); ++i)
		(*i)->calculateSalvage();

	for (RoidList::iterator i = m_Roids.begin(); i != m_Roids.end(); ++i)
	{
		for (vector<int>::iterator j = i->second.begin(); j != i->second.end(); ++j)
			cout << i->first << " roids : " << (*j) << endl;
	}
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::calcOneTick(Planet* friendly, Fleet* hostile, std::map<std::string, std::map<std::string, int> >& stealfriendly, std::map<std::string, std::map<std::string, int> >&  stealhostile, int tick)
{
	if (planetScore(tick - 1) > 0)
		setCapping(float(planetScore(tick - 1)) / hostile->score() /  10, tick);
	else
		setCapping(0, tick);
			
	//FIXME: remove this line later.. map<string, int> pods;


	//FIXME: Need to change this and allow multiple shiptypes with the same initiative. 
	//FIXME: Looks like I don't have to.. 
	map<int, string> unitsinit; // order units after their ininitiative
	for (UnitList::iterator i = s_Units.begin(); i != s_Units.end(); ++i)
		unitsinit[i->second.initiative()] = i->first;

	

	for (map<int, string>::iterator i = unitsinit.begin(); i != unitsinit.end(); ++i)
	{
		Fleet* hostiletemp = new Fleet(*hostile);
		Planet* friendlytemp = new Planet(*friendly);
		
		string unittype = i->second;
		int init = i->first;

		int freefriendly = friendly->freeFleet(unittype, 1);
		int freehostile = hostile->freeFleet(unittype, 1);

		if ( freefriendly <= 0 && freehostile <= 0)
			continue;

		//cerr << "Initiative: " << s_Units[unittype].initiative() << " with unit: " << unittype << endl;
	
		map<string, int> friendlyhits;
		map<string, int> hostilehits;

		if (s_Units[unittype].type() == "EMP")	
		{
			hostiletemp->takeEMP(unittype, freefriendly, friendlyhits);
			friendlytemp->takeEMP(unittype, freehostile, hostilehits);
		}
		else if (s_Units[unittype].type() == "Steal")
		{
			hostiletemp->takeShoot(unittype, freefriendly, stealfriendly[unittype]);
			friendlytemp->takeShoot(unittype, freehostile, stealhostile[unittype]);
			
			friendlyhits = stealfriendly[unittype];
			hostilehits = stealhostile[unittype];

			friendlytemp->calculateLostStealships(unittype, stealfriendly[unittype], 1);
			hostiletemp->calculateLostStealships(unittype, stealhostile[unittype], 1);
		}
		else
		{
			hostiletemp->takeShoot(unittype, freefriendly, friendlyhits);
			friendlytemp->takeShoot(unittype, freehostile, hostilehits);
		}


		if (s_Units[unittype].type() == "Pod")
		{
			//FIXME: Prolly need to recode the whole capping section for r7, due to multiple pods thingy
			cerr << "Capping is: " << capping(tick) << endl;
	
			if (capping(tick) > 0)
			{
				for (RoidList::iterator roid = m_Roids.begin(); roid != m_Roids.end(); ++roid)
				{
					int caproids = int(capping(tick) * roids(roid->first, tick - 1));
					//int freepods = hostiletemp->freeFleet(unittype, 1);
		
					cerr << "Possible to steal " << caproids << " " << roid->first << " roids\n";
					cerr << hostiletemp->freePodGuns(1) << " free pod guns available\n";
				
					if (hostiletemp->freePodGuns(1) <= 0)
						break;
					if (hostiletemp->freePodGuns(1) < caproids)
						caproids = hostiletemp->freePodGuns(1);
		
					cerr << caproids << " roids stolen\n";

					takeRoids(roid->first, caproids, tick);
		
					/*FIXME: REMOVE THIS BLOCK LATER
					//FIXME: Going to move this to the distribute roids section instead.. Not really move, I'll keep this, but "regenerate" the pods in the distribute ships function.
					hostiletemp->killFleet(unittype, caproids, 1);
					pods[unittype] += caproids;
					//int totroids = caproids + hostiletemp->resource(roid->first, 0);
					*/
					hostiletemp->addResource(roid->first, caproids, 1);
					hostiletemp->usePodGuns(1, caproids);
					cerr << caproids << " stolen " << roid->first << " roids\n";
				}
			}

			//FIXME: This is ugly and I want to change this, but as long as pods shoot last it works.. 
			*friendly = *friendlytemp;
			delete friendlytemp;
			*hostile = *hostiletemp;
			delete hostiletemp;

			return;
		}
		if (freefriendly > 0)
		{
			m_Report[tick][init]["Friendly"][unittype] = friendlyhits;
			m_Report[tick][init]["Friendly"][unittype]["000"] = freefriendly;
		}
		if (freehostile > 0)
		{
			m_Report[tick][init]["Hostile"][unittype] = hostilehits;
			m_Report[tick][init]["Hostile"][unittype]["000"] = freehostile;
		}

		//set the the objects so they point at the modified objects
		*friendly = *friendlytemp;
		delete friendlytemp;
		*hostile = *hostiletemp;
		delete hostiletemp;
	}
	//hostile->addFleet(pods, 1);
	
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::calculateScoreLoss(std::vector<Fleet*> friendly, int tick)
{
	for (vector<Fleet*>::iterator i = friendly.begin(); i != friendly.end(); ++i)
	{
		if ((*i)->name().find("Home") != string::npos)
		{
			int scorechange = (*i)->score(tick) - (*i)->score(tick - 1);
			addPlanetScore(scorechange, tick);
		}
	}
}

//////////////////////////////////////////////////////////////////////////
//
float Planet::capping(int tick)
{
	int ticks = m_Capping.size();

	--ticks;

	if (ticks < tick)
		return 0;

	return m_Capping[tick];
}

//////////////////////////////////////////////////////////////////////////
//
void Planet::setCapping(float capping, int tick)
{
	int ticks = m_Capping.size();

	for (int i = ticks; i <= tick; ++i)
	{
		m_Capping.push_back(0);
	}

	if (capping <= 0.15 && capping >= 0)
		m_Capping[tick] = capping;
	else if (capping >= 0.15)
		m_Capping[tick] = 0.15;
	else
		m_Capping[tick] = 0;
}

//////////////////////////////////////////////////////////////////////////
//
ReportList Planet::report() const
{
	return m_Report;
}
